#include <exception>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <functional>
#include <thread>
#include <future>
#include <mutex>

using namespace std;

static void printNestedException(const exception &ex, int level = 0)
{
    cerr << string(level, ' ').c_str() << "exception: " << ex.what() << '\n';
    try {
        rethrow_if_nested(ex);
    } catch (const exception &e) {
        printNestedException(e, level + 1);
    }
}

class Parser
{
public:
    void run(const string &fileName, function<void(void)> doneCallback);

    int getResult() const;
private:
    void runInternal(const string &fileName);
    void openFile(const string &fileName);

    atomic<int> m_result;
    exception_ptr m_exception;
    mutable mutex m_mutex;
};

void Parser::run(const string &fileName, function<void(void)> doneCallback)
{
    thread t1{[=]{
        try {
            runInternal(fileName);
        } catch (...) {
            lock_guard<mutex> guard{m_mutex};
            m_exception = current_exception();
        }
        doneCallback();
    }};
    t1.detach();
}

int Parser::getResult() const
{
    lock_guard<mutex> guard{m_mutex};
    if (m_exception)
        rethrow_exception(m_exception);
    return m_result;
}

void Parser::runInternal(const string &fileName)
{
    try {
        openFile(fileName);

        // ...

        lock_guard<mutex> guard{m_mutex};
        m_result = 42;
    } catch(...) {
        throw_with_nested( runtime_error("Parsing failed!") );
    }
}

void Parser::openFile(const string &fileName)
{
    try {
        ifstream file(fileName);
        file.exceptions(ios_base::failbit);
    } catch(...) {
        throw_with_nested( runtime_error("Couldn't open " + fileName) );
    }
}

promise<void> prom;

int main()
{
    auto fut = prom.get_future();

    // Start the parser, which will spawn a seperate thread.
    // Once the parser is finished, it will call the lambda function.
    // For the sake of the example, we use future and promise to simply
    // let the main thread wait until the parser thread has finished.

    Parser parser;
    parser.run("data.txt", [] {
        prom.set_value();
    });

    fut.get();

    try {
        cout << parser.getResult() << endl;
    } catch (exception &ex) {
        printNestedException(ex);
    }

    return 0;
}
