#include <utility>
#include <typeinfo>
#include <iostream>
#include <string>
#include <type_traits>

#ifdef _MSC_VER
#pragma warning( disable:4100 )
#endif

using namespace std;

class TestType {
public:
    TestType() {
        cout << "Test constructor\n";
    }

    // The compiler often elides the copy and move constructors, so you won't see them
    // printed as often as you would think. This is a compiler optimization.
    TestType(const TestType &) {
        cout << "Test copy constructor\n";
    }

    TestType(TestType &&) {
        cout << "Test move constructor\n";
    }
};

void foo(const TestType &)
{
    cout << "foo overload for const lvalue reference" << endl;
}

void foo(TestType &&)
{
    cout << "foo overload for rvalue reference" << endl;
}

void foo(const TestType &&)
{
    cout << "foo overload for const rvalue reference" << endl;
}

void foo(TestType &)
{
    cout << "foo overload for lvalue reference" << endl;
}

// Universal reference - we don't need an overload for all 4 different types, nice!
template<typename T>
void printTypeInfo(const string &message, T &&t)
{
    // We use decltype(t) instead of T here, as otherwise the info on the reference type would be missing

    string typeDescription;
    if (is_const<typename remove_reference<T>::type>::value)
        typeDescription += "const ";
    if (is_rvalue_reference<decltype(t)>::value)
        typeDescription += "rvalue reference to ";
    if (is_lvalue_reference<decltype(t)>::value)
        typeDescription += "lvalue reference to ";
    typeDescription += typeid(t).name();
    cout << message << ' ' << typeDescription << endl;
}

// Notice how adding "const" to the parameter here will trigger compilation failure, as that disables
// reference collapsing - "const T&&" is always a const rvalue reference, whereas "T &&" is a universal
// reference to which reference collapsing rules apply.
// If the parameter is "const T&&", then this function can only be called with const rvalue references,
// i.e. const temporaries, and therefore "forwardTest(lvalue)" will trigger compilation failure.
//
// "const" is not needed here anyway, as the "const" is carried over as part of the type T.
template<typename T>
void forwardTest(/*const*/ T&& t)
{
    printTypeInfo("forwardTest called with type", forward<T>(t));

    // Technically, std::forward() can only be used once for a local variable. Once something is
    // forwarded, the local variable might not be valid anymore, as the code that was passed
    // the forwarded object might have moved, and therefore invalidated, it.
    // This is the same as with std::move(), which also invalidates an object.
    //
    // So technically, we are not allowed to use "t" here, as printTypeInfo() might have
    // invalidated it.
    // We know however that printTypeInfo() does not do anything to "t", so we ignore that
    // for example's sake.
    //
    // In real code, never use a variable after it has been forwarded or moved.
    foo(forward<T>(t));

    // Same as the above. But always use forward() in your code, it makes the intention clear!
    //foo(static_cast<T&&>(t));

    // t is an lvalue! Therefore the commented out version below would always call the foo() overload
    // for lvalues and lvalue references. Hence the call to forward() above is needed.
    //foo(t);
}

TestType rvalue() {
    return TestType();
}

const TestType constRvalue() {
    return TestType();
}

int main()
{
   TestType lvalue;
   const TestType constLvalue;

   cout << endl;
   cout << "--- Now calling with rvalue ---" << endl;
   forwardTest(rvalue());
   cout << endl;

   cout << "--- Now calling with lvalue ---" << endl;
   forwardTest(lvalue);
   cout << endl;

   cout << "--- Now calling with const rvalue ---" << endl;
   forwardTest(constRvalue());
   cout << endl;

   cout << "--- Now calling with const lvalue ---" << endl;
   forwardTest(constLvalue);
}
