#include <random>
#include <iostream>
#include <array>
#include <vector>
#include <string>
#include <algorithm>
#include <iomanip>
#include <cassert>

using namespace std;

template<class Distribution, class Generator>
void printDistribution(Distribution dist, Generator gen,
                       int min, int max)
{
    // Calculate some properties of the histogram
    assert(max > min);
    int numBuckets = 20;
    static const int numSamples = 10000;
    static const int displayWidth = 80;
    const int range = max - min + 1;
    if (range < numBuckets)
        numBuckets = range;
    const int bucketSize = range / numBuckets;
    
    // Get some random samples to generate the histogram
    std::vector<int> histogram(numBuckets, 0);
    for (int n = 0; n < numSamples; ++n) {
        const double sample = dist(gen);
        if (sample >= min && sample <= max) {
            const int index = static_cast<int>((sample - min) / bucketSize);
            ++histogram[index];
        }
    }

    cout << endl << endl;

    // Print the histogram
    const int maxSamples = *max_element(begin(histogram), end(histogram));
    const float scaleFactor = maxSamples / static_cast<float>(displayWidth);
    for (int i = 0; i < numBuckets; i++) {
        const int bucketMean = i * bucketSize + min + bucketSize / 2;
        cout << setw(4) << bucketMean << " ";
        cout << string(static_cast<int>(histogram[i] / scaleFactor), '*');
        cout << " (" << histogram[i] << ")" << endl;
    }
}

int main()
{
    // mt19937 is a good default generator, use that unless random number generation needs to be faster.
    //
    // We seed the mt19937 generator with a random number from random_device, which uses the OS's entropy
    // source, and therefore produces 'real' random numbers, whereas the other generators produce pseudo-random
    // numbers.
    // random_device should be used as sparingly as possible, as it drains the OS's entropy pool.
    random_device random;
    mt19937 generator{random()};

    const int randomNumber = generator();
    cout << "Random number from the mersenne twister: " << randomNumber << endl;

    uniform_int_distribution<int> diceDistribution{1, 6};
    const int diceRoll = diceDistribution(generator);
    cout << "Dice roll sample: " << diceRoll << endl;

    normal_distribution<double> iqDistribution{100, 15};
    const int iq = static_cast<int>(iqDistribution(generator));
    cout << "IQ sample: " << iq << endl;

    printDistribution(diceDistribution, generator,
                      diceDistribution.min(), diceDistribution.max());
    printDistribution(iqDistribution, generator,
                      static_cast<int>(iqDistribution.mean() - iqDistribution.stddev() * 4),
                      static_cast<int>(iqDistribution.mean() + iqDistribution.stddev() * 4));
}

