//
// C++98 approach: Structs that are specialized for a type, static functions
// that return the value. This is the approach used by std::numeric_limits,
// although without the constexpr in C++98-mode.
//

template<typename T>
struct numeric_limits_98 {
    static constexpr T max() {
        return T();
    }
};

template<>
struct numeric_limits_98<int> {
    static constexpr int max() {
        return 640'000; // 640k ought to be enough for everyone
    }
};

template<>
struct numeric_limits_98<float> {
    static constexpr float max() {
        return 999.99f; // Biggest number I can think of
    }
};

//
// C++11 approach: Structs that are specialized for a type, static constexpr member
// variables that contain the value.
//

template<typename T>
struct numeric_limits_11 {
    constexpr static T max = T();
};

template<>
struct numeric_limits_11<int> {
    constexpr static int max = 640'000;
};

template<>
struct numeric_limits_11<float> {
    constexpr static float max = 999.99f;
};

template<typename T>
constexpr T numeric_limits_11<T>::max;

//
// C++14 approach: Variable templates
//

namespace numeric_limits_14 {
    template<typename T>
    constexpr T max = T();

    template<>
    constexpr int max<int> = 640'000;

    template<>
    constexpr float max<float> = 999.99f;
}

static_assert(numeric_limits_98<int>::max() == 640'000, "Fail");
static_assert(numeric_limits_98<float>::max() == 999.99f, "Fail");

static_assert(numeric_limits_11<int>::max == 640'000, "Fail");
static_assert(numeric_limits_11<float>::max == 999.99f, "Fail");

static_assert(numeric_limits_14::max<int> == 640'000, "Fail");
static_assert(numeric_limits_14::max<float> == 999.99f, "Fail");

int main(int , char *[])
{
}
