#include <algorithm>
#include <cmath>
#include <iostream>
#include <memory>
#include <numeric>
#include <string>
#include <vector>

using namespace std;

//
// Please follow the proposed steps in this file.
// They are marked with "STEP X:".
//
// Overview:
//  * Gradually replace the concrete source node types with a function
//    configurable alternative
//  * Gradually replace the concrete process node types with a function
//    configurable alternative
//  * Gradually replace the concrete sink node types with a function
//    configurable alternative
//  * Remove the unused node types
//  * Simplify the resulting code
//

////////////////////////////////////////////////////////////////
/// Functions provided for convenience
////////////////////////////////////////////////////////////////

static ostream &operator <<(ostream &stream, const vector<long long> &values)
{
    string content = accumulate(begin(values), end(values), string(),
                                [] (const string &content, long long value) {
                                    if (content.empty())
                                        return to_string(value);
                                    else
                                        return content + ", " + to_string(value);
                                });
    stream << "(" << content << ")";
    return stream;
}

static bool isPrime(long long n)
{
    if (n <= 1)
        return false;

    for (long long i = 2; i < n; i++) {
        if (n % i == 0)
            return false;
    }

    return true;
}

static long long nthPrime(long long n)
{
    long long numPrimesFound = 0;
    for (long long i = 1;; i++) {
        if (isPrime(i)) {
            numPrimesFound++;
            if (numPrimesFound == n)
                return i;
        }
    }
}

static long long fib(long long n)
{
    if (n <= 2)
        return 1;
    else
        return fib(n - 1) + fib(n - 2);
}

static long long powLL(long long base, int exp)
{
    return static_cast<long long>(pow(static_cast<long double>(base),
                                      static_cast<long double>(exp)));
}


////////////////////////////////////////////////////////////////
/// Pipeline interfaces
////////////////////////////////////////////////////////////////

class SourceNodeBase
{
public:
    virtual ~SourceNodeBase() = default;
    virtual long long generateValue() = 0;
};

class SinkNodeBase
{
public:
    virtual ~SinkNodeBase() = default;
    virtual void consumeValue(long long value) = 0;
};

class ProcessNodeBase
{
public:
    virtual ~ProcessNodeBase() = default;
    virtual long long processValue(long long value) = 0;
};

class Pipeline
{
public:
    Pipeline(const shared_ptr<SourceNodeBase> &source,
             const vector<shared_ptr<ProcessNodeBase>> &processing,
             const shared_ptr<SinkNodeBase> &sink)
        : m_sourceNode(source),
          m_processNodes(processing),
          m_sinkNode(sink)
    {
    }

    void processOne()
    {
        long long value = m_sourceNode->generateValue();
        for (auto node : m_processNodes)
            value = node->processValue(value);
        m_sinkNode->consumeValue(value);
    }

private:
    shared_ptr<SourceNodeBase> m_sourceNode;
    vector<shared_ptr<ProcessNodeBase>> m_processNodes;
    shared_ptr<SinkNodeBase> m_sinkNode;
};



////////////////////////////////////////////////////////////////
/// Concrete pipeline nodes
////////////////////////////////////////////////////////////////

// STEP 1: Add a concrete SourceNode inheriting from SourceNodeBase
// it should be creatable with any function returning a long long.

class OddSource : public SourceNodeBase
{
public:
    long long generateValue() override
    {
        const long long value = m_nextValue;
        m_nextValue += 2;
        return value;
    }

private:
    long long m_nextValue = 1;
};

class EvenSource : public SourceNodeBase
{
public:
    long long generateValue() override
    {
        const long long value = m_nextValue;
        m_nextValue += 2;
        return value;
    }

private:
    long long m_nextValue = 2;
};


// STEP 3: Add a concrete ProcessNode inheriting from ProcessNodeBase
// it should be creatable with any function taking a long long as parameter
// and returning a long long.

class FibNode : public ProcessNodeBase
{
public:
    long long processValue(long long value) override
    {
        return fib(value);
    }
};

class PrimeNode : public ProcessNodeBase
{
public:
    long long processValue(long long value) override
    {
        return nthPrime(value);
    }
};

class Pow3Node : public ProcessNodeBase
{
public:
    long long processValue(long long value) override
    {
        return powLL(value, 3);
    }
};

class Mul2Node : public ProcessNodeBase
{
public:
    long long processValue(long long value) override
    {
        return value * 2;
    }
};

// STEP 5: Add a concrete SinkNode inheriting from SinkNodeBase
// it should be creatable with any function taking a long long as parameter.

class VectorSink : public SinkNodeBase
{
public:
    void consumeValue(long long value) override
    {
        values.push_back(value);
    }

    vector<long long> values;
};

class SumSink : public SinkNodeBase
{
public:
    void consumeValue(long long value) override
    {
        sum += value;
    }

    long long sum = 0;
};

// STEP 7: Remove all the old concrete nodes implementation which
// are now unused.

// STEP 8: Look for more simplification opportunities

////////////////////////////////////////////////////////////////
/// Test pipelines
////////////////////////////////////////////////////////////////

int main()
{
    // STEP 2: Remove OddSource and EvenSource uses, replace them
    // with SourceNode instances.

    // STEP 4: Remove PrimeNode, Mul2Node, FibNode and Pow3Node uses,
    // replace them with ProcessNode instances.

    // STEP 6: Remove VectorSink and SumSink uses, replace them
    // with SinkNode instances.

    auto pipe1Sink = new SumSink;
    auto pipe1 = Pipeline(make_shared<EvenSource>(),
                          {make_shared<PrimeNode>(), make_shared<Mul2Node>()},
                          shared_ptr<SinkNodeBase>(pipe1Sink));

    auto pipe2Sink = new VectorSink;
    auto pipe2 = Pipeline(make_shared<OddSource>(),
                          {make_shared<FibNode>(), make_shared<Pow3Node>()},
                          shared_ptr<SinkNodeBase>(pipe2Sink));

    for (int i = 0; i < 7; i++) {
        pipe1.processOne();
        pipe2.processOne();
    }

    cout << "Pipeline 1 result: " << pipe1Sink->sum << endl;
    cout << "Pipeline 2 result: " << pipe2Sink->values << endl;

    return 0;
}
