#include <iostream>
#include <vector>
#include <string>
#include <algorithm>
#include <functional>
#include <iterator>

using namespace std;

//
// For brevity, everything was made public in the two classes.
//

class Person
{
public:
    string name;

    explicit Person(const string &n)
        : name(n)
    {
    }

    virtual ~Person() = default;

    virtual void printInfo() const {
        cout << name << endl;
    }
};

class Employee : public Person
{
public:

    // C++11 change: Using comma at end of list
    enum Job {
        Sysadmin,
        SoftwareEngineer,
        Manager,
    };

    // C++11 change: Using nullptr
    Employee(const string &name, Job job, int salary)
        : Person{name}, job{job}, salary{salary}, boss{nullptr}
    {
        id = highestEmployeeId++;
    }

    Job job;
    int salary;
    Employee *boss;
    int id;
    static int highestEmployeeId;

    // C++11 change: Using override to better detect errors
    void printInfo() const override {
        Person::printInfo();
        cout << "Id: " << id << ", Salary: " << salary << endl << endl;
    }
};

int Employee::highestEmployeeId = 1;

int main()
{
    // C++11 change: Using brace initializer for Employee
    // C++11 change: Using raw string literals
    // C++11 change: Use initalizer_list constructor of vector
    vector<Employee> employees = {
        {R"(Frank "Hitman" Hennessy)",
         Employee::Sysadmin, 40000},
        {R"(Earl "Magic" Walker)",
         Employee::SoftwareEngineer, 45000},
        {R"(Walter "Heisenberg" White)",
         Employee::Manager, 50000},
        {R"(Peter "Wolf" Sanderson)",
         Employee::SoftwareEngineer, 39000}
    };

    // C++11 change: Using auto and range-based for loops
    for (const auto &employee : employees) {
        employee.printInfo();
    }

    // Calculate how many people work in IT
    // C++11 change: Using cbegin()/cend() and using a lambda
    const auto itCount = count_if(employees.cbegin(), employees.cend(),
                                  [](const Employee &employee) {
        return employee.job == Employee::SoftwareEngineer ||
               employee.job == Employee::Sysadmin;
    });
    cout << "A total of " << itCount << " people work in IT." << endl;

    // Find highest earner and print his name
    // C++11 change: Using a lambda and auto
    auto highestEarner = max_element(employees.begin(),
           employees.end(), [](const Employee &e1, const Employee &e2) {
        return e1.salary < e2.salary;
    });
    cout << "Highest earner in the company is " << highestEarner->name << endl << endl;

    // Now fire him because he is costing the company so much!
    employees.erase(highestEarner);

    // Give all other employees a raise of 10%!
    // C++11 change: Using lambda and std::transform()
    vector<Employee> newEmployees;
    transform(employees.cbegin(), employees.cend(), back_inserter(newEmployees),
              [](const Employee& employee) {
        Employee result{employee};
        result.salary = static_cast<int>(result.salary * 1.1f);
        return result;
    });

    // C++11 change: Using for_each with mem_fn
    cout << "Old Salaries:" << endl;

    // mem_fn triggers https://llvm.org/bugs/show_bug.cgi?id=21026 on some machines,
    // so use a lambda instead.
    for_each(employees.cbegin(), employees.cend(), [] (const Employee &e) {
        e.printInfo();
    });
    //for_each(employees.cbegin(), employees.cend(),
    //         mem_fn(&Employee::printInfo));

    cout << endl << "New Salaries:" << endl;
    //for_each(newEmployees.cbegin(), newEmployees.cend(),
    //         mem_fn(&Employee::printInfo));
    for_each(newEmployees.cbegin(), newEmployees.cend(), [] (const Employee &e) {
        e.printInfo();
    });
}
