#include <iostream>
#include <cassert>

// tolower() from the standard library is not constexpr, so provide our own,
// not locale-aware, version.
static constexpr char tolower(char c)
{
    if (c >= 'A' && c <= 'Z')
        return c - ('A' - 'a');
    else
        return c;
}

// Compile-time "unit test" of tolower()
static_assert(tolower('A') == 'a', "tolower() failure!");
static_assert(tolower('a') == 'a', "tolower() failure!");
static_assert(tolower('Z') == 'z', "tolower() failure!");
static_assert(tolower('z') == 'z', "tolower() failure!");
static_assert(tolower('!') == '!', "tolower() failure!");

class String
{
public:
    constexpr String() {}

    constexpr String(const char *data)
        : m_data(data)
    {
    }

    // Another possible constructor, which would take away the need to calculate the size
    /*
    template<int N>
    constexpr String(const char(&data)[N])
        : m_data(data), m_size(N)
    {
    }
    */

    constexpr int size() const
    {
        if (!m_data)
            return 0;

        int i = 0;
        while (m_data[i] != 0)
            i++;
        return i;
    }

    constexpr int vowelCount() const
    {
        if (!m_data)
            return 0;

        int count = 0;
        for (int i = 0; i < size(); i++) {
            char c = tolower(m_data[i]);
            if (c == 'a' || c == 'e' || c == 'i' || c == 'o' || c == 'u')
                count++;
        }
        return count;
    }

    // constexpr, but not const
    constexpr void clear()
    {
        m_data = nullptr;
    }

    // const, but not constexpr
    void dumpInfo() const
    {
        std::cout << "Info for string '" << *this << "':" << std::endl;
        std::cout << "  size: " << size() << std::endl;
        std::cout << "  vowels: " << vowelCount() << std::endl;
    }

    constexpr operator const char*() const
    {
        return m_data;
    }

    constexpr char operator[](int index) const
    {
        if (index < 0 || index >= size())
            throw std::out_of_range("Invalid array index!");
        return m_data[index];
    }

private:
    const char *m_data = nullptr;
};

static constexpr String createEmptyString()
{
    String result("nonsense");
    result.clear();
    return result;
}

int main()
{
    constexpr String helloWorld("Hello World! Adios!");
    static_assert(helloWorld.size() == 19, "size() broken!");
    static_assert(helloWorld.vowelCount() == 6, "vowelCount() broken!");
    static_assert(helloWorld[0] == 'H', "operator[] broken!");
    static_assert(helloWorld[4] == 'o', "operator[] broken!");
    try {
        helloWorld[4711];
        assert(false); // Unreachable
    } catch (const std::out_of_range &) {
    }
    helloWorld.dumpInfo();

    constexpr String test("constexpr is fast!");
    static_assert(test.size() == 18, "size() broken!");
    static_assert(test.vowelCount() == 4, "vowelCount() broken!");
    std::cout << test << std::endl;

    constexpr String null;
    static_assert(null.size() == 0, "size() broken!");
    static_assert(null.vowelCount() == 0, "vowelCount() broken!");

    constexpr String empty("");
    static_assert(empty.size() == 0, "size() broken!");
    static_assert(empty.vowelCount() == 0, "vowelCount() broken!");

    constexpr String empty2 = createEmptyString();
    static_assert(empty2.size() == 0, "size() broken!");
    static_assert(empty2.vowelCount() == 0, "vowelCount() broken!");

    constexpr int arraySize = 128;
    char randomChars[arraySize];
    randomChars[arraySize - 1] = 0;
    for (int i = 0; i < arraySize - 1; i++) {
        randomChars[i] = 'A' + rand() % 24;
    }
    const String random(randomChars);
    assert(random.size() == arraySize - 1);
    random.dumpInfo();
}
