#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <functional>
#include <stdexcept>
#include <type_traits>
#include <utility>
#include <chrono>

#ifdef _MSC_VER
#   if _MSC_VER <= 1800
#       define KDAB_NOEXCEPT
#   else
#       define KDAB_NOEXCEPT noexcept
#   endif
#else
#    define KDAB_NOEXCEPT noexcept
#endif


using namespace std;

// There already is a typedef for chrono::milliseconds, however that one
// uses integer representation, whereas we need floating point representation
// for more accuracy.
using ms = chrono::duration<double, milli>;

static int numCopyCtorsCalled = 0;

class BigObject
{
    enum { DataSize = 100000 };
public:

    BigObject()
        : data{new int[DataSize]}
    {
        //cout << "BigObject constructor" << endl;
        std::generate(data, data + DataSize, rand);
    }

    ~BigObject() KDAB_NOEXCEPT
    {
        delete[] data;
        data = nullptr;
    }

    BigObject(const BigObject &other)
        : data{new int[DataSize]}
    {
        //cout << "BigObject copy constructor" << endl;
        numCopyCtorsCalled++;
        copy(other.data, other.data + DataSize, data);
    }

    BigObject(BigObject &&other) KDAB_NOEXCEPT
    {
        //cout << "BigObject move constructor" << endl;
        data = other.data;
        other.data = nullptr;
    }

    void changeSomething()
    {
        data[rand() % DataSize] = rand();
    }

    void printContent() const
    {
        for_each(data, data + DataSize, [](int /*value*/) {
            //cout << value << ",";
        });
    }

private:
    int *data;
};

static BigObject createBigObject()
{
    BigObject foo;
    BigObject bar;

    // Two return statements to prevent NRVO optimization
    if (rand() == 0)
        return foo;
    return bar;
}

static BigObject createAnotherBigObject()
{
    /* const */ BigObject foo;
    /* const */ BigObject bar;

    if (rand() == 0)
        return foo;
    return bar;
}

static BigObject createYetAnotherObject()
{
    return BigObject{};
}

// This overload is not needed for this specific solution and just produces
// a compiler warning about an unused function.
//
// In the usual case however, you would have two overloaded functions,
// one version with a const ref parameter, and one with a rvalue ref
// parameter.
/*
static void workWithBigObject(const BigObject &object)
{
    BigObject copy{object};
    copy.changeSomething();
    copy.printContent();
}
*/

static void workWithBigObject(BigObject &&object)
{
    // A move is still required here, as "object" is an lvalue(!)
    BigObject copy{move(object)};
    copy.changeSomething();
    copy.printContent();
}

static void doVectorStuff()
{
    vector<BigObject> bigObjects;    

    // Do not use reserve here, as that would be against the point of this test

    for (int i = 0; i < 5; i++) {
        // By using emplace_back(), we save one move constructor call, as the
        // object is constructed in-place instead of being constructed in here
        // and then moved.
        bigObjects.emplace_back();
    }
}

static void testObjectReturn()
{
    // This test simply need a move constructor in BigObject
    createBigObject();
}

static void testObjectReturn2()
{
    // In this test, remove const from the BigObject instances, as that prevents
    // move semantics (you can't steal the guts of a const object)
    createAnotherBigObject();
}

static void useTempObjectInFunction()
{
    // Use std::move() since we know "temp" is temporary, so it is ok if
    // workWithBigObject() steals its guts.
    BigObject temp;
    workWithBigObject(move(temp));
}

static void vectorResize()
{
    // This test requires that the destructor and the move constructor are
    // marked noexcept, as vector uses move_if_noexcept internally
    doVectorStuff();
}

static void useTempObjectInFunction2()
{
    // This test required removing the "const" from the return type of
    // createYetAnotherObject().
    // Even though using const in the return type was recommended in
    // Effective C++, 3rd edition, this no longer applies to C++11 code.
    workWithBigObject(createYetAnotherObject());
}

static void executeTest(std::function<void()> test, const string &name)
{
    cout << endl << endl;
    cout << "--------- Executing " << name << " -------------------" << endl;
    const int startCopyCtors = numCopyCtorsCalled;
    const auto start = chrono::high_resolution_clock::now();

    test();

    const auto elapsed = chrono::high_resolution_clock::now() - start;
    cout << name << " took" << ": " << ms{elapsed}.count() << " ms\n";
    cout << "Number of copy constructors called: "
         << numCopyCtorsCalled - startCopyCtors << endl;
}

int main()
{
    const auto start = chrono::high_resolution_clock::now();

    executeTest(testObjectReturn, "Object Return Test");
    executeTest(testObjectReturn2, "Object Return Test 2");
    executeTest(useTempObjectInFunction, "Using Object in Function Test");
    executeTest(vectorResize, "Resizing Vector Test");
    executeTest(useTempObjectInFunction2, "Using Object in Function Test 2");

    const auto elapsed = chrono::high_resolution_clock::now() - start;
    cout << endl << endl;
    cout << "All tests took" << ": " << ms{elapsed}.count() << " ms\n";
    cout << "Number of copy constructors called: " << numCopyCtorsCalled << endl;
}
