#include <algorithm>
#include <cmath>
#include <functional>
#include <iostream>
#include <memory>
#include <numeric>
#include <string>
#include <vector>

using namespace std;
using namespace std::placeholders;


////////////////////////////////////////////////////////////////
/// Functions provided for convenience
////////////////////////////////////////////////////////////////

static ostream &operator <<(ostream &stream, const vector<long long> &values)
{
    string content = accumulate(begin(values), end(values), string(),
                                [] (const string &content, long long value) {
                                    if (content.empty())
                                        return to_string(value);
                                    else
                                        return content + ", " + to_string(value);
                                });
    stream << "(" << content << ")";
    return stream;
}

static bool isPrime(long long n)
{
    if (n <= 1)
        return false;

    for (long long i = 2; i < n; i++) {
        if (n % i == 0)
            return false;
    }

    return true;
}

static long long nthPrime(long long n)
{
    long long numPrimesFound = 0;
    for (long long i = 1;; i++) {
        if (isPrime(i)) {
            numPrimesFound++;
            if (numPrimesFound == n)
                return i;
        }
    }
}

static long long fib(long long n)
{
    if (n <= 2)
        return 1;
    else
        return fib(n - 1) + fib(n - 2);
}

static long long powLL(long long base, int exp)
{
    return static_cast<long long>(pow(static_cast<long double>(base),
                                      static_cast<long double>(exp)));
}



////////////////////////////////////////////////////////////////
/// Pipeline classes
////////////////////////////////////////////////////////////////

class SourceNode
{
public:
    SourceNode(function<long long()> generator)
        : m_generator(generator) {}

    long long generateValue() { return m_generator(); }

private:
    function<long long()> m_generator;
};

class SinkNode
{
public:
    SinkNode(function<void(long long)> consumer)
        : m_consumer(consumer) {}

    void consumeValue(long long value) { m_consumer(value); }

private:
    function<void(long long)> m_consumer;
};

class ProcessNode
{
public:
    ProcessNode(function<long long(long long)> processor)
        : m_processor(processor) {}

    long long processValue(long long value) { return m_processor(value); }

private:
    function<long long(long long)> m_processor;
};

class Pipeline
{
public:
    Pipeline(shared_ptr<SourceNode> source,
             vector<shared_ptr<ProcessNode>> processing,
             shared_ptr<SinkNode> sink)
        : m_sourceNode(source),
          m_processNodes(processing),
          m_sinkNode(sink)
    {
    }

    void processOne()
    {
        long long value = m_sourceNode->generateValue();
        for (auto node : m_processNodes)
            value = node->processValue(value);
        m_sinkNode->consumeValue(value);
    }

private:
    shared_ptr<SourceNode> m_sourceNode;
    vector<shared_ptr<ProcessNode>> m_processNodes;
    shared_ptr<SinkNode> m_sinkNode;
};



////////////////////////////////////////////////////////////////
/// Concrete pipeline nodes
////////////////////////////////////////////////////////////////

shared_ptr<SourceNode> counterSource(long long init, long long step)
{
    long long nextValue = init;
    return make_shared<SourceNode>([=] () mutable {
        const long long value = nextValue;
        nextValue += step;
        return value;
    });
}

shared_ptr<SourceNode> oddSource()
{
    return counterSource(1, 2);
}

shared_ptr<SourceNode> evenSource()
{
    return counterSource(2, 2);
}

shared_ptr<ProcessNode> fibNode()
{
    return make_shared<ProcessNode>(fib);
}

shared_ptr<ProcessNode> primeNode()
{
    return make_shared<ProcessNode>(nthPrime);
}

shared_ptr<ProcessNode> powNode(int power)
{
    return make_shared<ProcessNode>(bind(powLL, _1, power));
}

shared_ptr<ProcessNode> mulNode(int factor)
{
    return make_shared<ProcessNode>([=] (long long value) { return value * factor; });
}

shared_ptr<SinkNode> vectorSink(vector<long long> &values)
{
    return make_shared<SinkNode>([&] (long long value) { values.push_back(value); });
}

shared_ptr<SinkNode> sumSink(long long &sum)
{
    return make_shared<SinkNode>([&] (long long value) { sum += value; });
}



////////////////////////////////////////////////////////////////
/// Test pipelines
////////////////////////////////////////////////////////////////

int main()
{
    auto sum = 0LL;
    auto pipe1 = Pipeline(evenSource(),
                          {primeNode(), mulNode(2)},
                          sumSink(sum));

    auto values = vector<long long>();
    auto pipe2 = Pipeline(oddSource(),
                          {fibNode(), powNode(3)},
                          vectorSink(values));

    for (int i = 0; i < 7; i++) {
        pipe1.processOne();
        pipe2.processOne();
    }

    cout << "Pipeline 1 result: " << sum << endl;
    cout << "Pipeline 2 result: " << values << endl;

    return 0;
}
