#include <iostream>
#include <list>
#include <functional>
#include <future>
#include <mutex>

using namespace std;

using ulonglong = unsigned long long;

bool isPrime(ulonglong n)
{
    if (n <= 1)
        return false;

    for (ulonglong i = 2; i < n; i++) {
        if (n % i == 0)
            return false;
    }

    return true;
}

ulonglong nthPrime_impl(ulonglong n)
{
    ulonglong numPrimesFound = 0;
    for (ulonglong i = 1;; i++) {
        if (isPrime(i)) {
            numPrimesFound++;
            if (numPrimesFound == n)
                return i;
        }
    }
}

static ulonglong fib_impl(ulonglong n)
{
    if (n <= 2)
        return 1;
    else
        return fib_impl(n - 1) + fib_impl(n - 2);
}

class Calculator
{
public:
    void run();
    void stop();

    future<ulonglong> nthPrime(ulonglong n);
    future<ulonglong> fib(ulonglong n);

private:
    future<ulonglong> createTask(ulonglong n,
                                 function<ulonglong(ulonglong)> func);

    struct Task {
// MSVC <= 2013 doesn't implicitly generate move constructors, so do that ourselves
#if defined(_MSC_VER ) && _MSC_VER <= 1800
        Task() = default;
        Task(Task &&other)
            : prom(move( other.prom ))
            , func(move( other.func ))
            , input(other.input)
        {}
#endif
        promise<ulonglong> prom;
        function<ulonglong(ulonglong)> func;
        ulonglong input;
    };

    list<Task> m_tasks;
    mutex m_mutex;
    condition_variable m_condition;
    bool m_cancel = false;
};

future<ulonglong> Calculator::createTask(
        ulonglong n, function<ulonglong(ulonglong)> func)
{
    Task task;
    task.input = n;
    task.func = func;
    future<ulonglong> fut = task.prom.get_future();

    {
        const unique_lock<mutex> lock{m_mutex};
        m_tasks.push_back(move(task));
    }

    m_condition.notify_one();

    return fut;
}

future<ulonglong> Calculator::nthPrime(ulonglong n)
{
    return createTask(n, nthPrime_impl);
}

future<ulonglong> Calculator::fib(ulonglong n)
{
    return createTask(n, fib_impl);
}

void Calculator::run()
{
    unique_lock<mutex> lock{m_mutex};
    while (!m_cancel) {
        m_condition.wait(lock, [this] {
            return !m_tasks.empty() || m_cancel;
        });

        while (!m_tasks.empty() && !m_cancel) {
            Task task = move(m_tasks.front());
            m_tasks.pop_front();

            // Unlock mutex while doing the calculation, so that
            // other threads don't block when calling nthPrime()
            // or fib().
            lock.unlock();
            const ulonglong result = task.func(task.input);
            lock.lock();

            task.prom.set_value(result);
        }
    }
}

void Calculator::stop()
{
    const unique_lock<mutex> lock{m_mutex};
    m_cancel = true;
    m_condition.notify_one();
}

int main()
{
    Calculator calc;
    thread t{&Calculator::run, &calc};
    auto prime = calc.nthPrime(5000);
    auto fib = calc.fib(41);

    cout << "The prime is: " << prime.get() << endl;
    cout << "The fib is: " << fib.get() << endl;

    // Make sure to properly stop the running thread
    calc.stop();
    t.join();
}
