#include <algorithm>
#include <array>
#include <chrono>
#include <iostream>
#include <string>
#include <thread>

using namespace std;


////////////////////////////////////////////////////////////////
/// Test harness, for reading purpose only
////////////////////////////////////////////////////////////////

using ms = chrono::duration<double, milli>;
using ulonglong = unsigned long long;
static const int ARRAY_SIZE = 10;
static const ulonglong RUN_COUNT = 1000000;

template<typename T>
static std::ptrdiff_t switchesStillOnCount(const array<T, ARRAY_SIZE> &boxes)
{
    // Not using the mem_fn solution below because of https://llvm.org/bugs/show_bug.cgi?id=21026
    auto isFirstSwitchOn = [] (const T& t) { return t.isFirstSwitchOn(); };
    auto isSecondSwitchOn = [] (const T& t) { return t.isSecondSwitchOn(); };

    return count_if(begin(boxes), end(boxes), isFirstSwitchOn)
         + count_if(begin(boxes), end(boxes), isSecondSwitchOn);

    //return count_if(begin(boxes), end(boxes), mem_fn(&T::isFirstSwitchOn))
    //     + count_if(begin(boxes), end(boxes), mem_fn(&T::isSecondSwitchOn));
}

template<typename T>
static void toggleFirstSwitchesSequential(array<T, ARRAY_SIZE> &boxes)
{
    for (ulonglong run = 0; run < 2 * RUN_COUNT + 1; run++) {
        for (auto &box : boxes)
            box.toggleFirstSwitch();
    }
}

template<typename T>
static void toggleSecondSwitchesSequential(array<T, ARRAY_SIZE> &boxes)
{
    for (ulonglong run = 0; run < 2 * RUN_COUNT + 1; run++) {
        for (auto &box : boxes)
            box.toggleSecondSwitch();
    }
}

template<typename T>
static void executeTest(std::function<void(array<T, ARRAY_SIZE> &)> switcher, const string &name)
{
    cout << endl << endl;
    cout << "--------- Executing " << name << " -------------------" << endl;
    const auto start = chrono::high_resolution_clock::now();

    array<T, ARRAY_SIZE> boxes;
    switcher(boxes);

    const auto elapsed = chrono::high_resolution_clock::now() - start;
    cout << "Test took: " << ms{elapsed}.count() << " ms" << endl;
    cout << "Size of each object is: " << sizeof(T) << " bytes" << endl;
    cout << "Number of switches still on: " << switchesStillOnCount(boxes) << endl;
}



////////////////////////////////////////////////////////////////
/// Sequential algorithm for baseline evaluation
/// Do not modify
////////////////////////////////////////////////////////////////
template<typename T>
static void toggleAllSwitchesSequential(array<T, ARRAY_SIZE> &boxes)
{
    toggleFirstSwitchesSequential(boxes);
    toggleSecondSwitchesSequential(boxes);
}


////////////////////////////////////////////////////////////////
/// Switching the two types of switches should be done in parallel
/// Please modify
////////////////////////////////////////////////////////////////
template<typename T>
static void toggleAllSwitchesParallel(array<T, ARRAY_SIZE> &boxes)
{
    thread t{toggleFirstSwitchesSequential<T>, ref(boxes)};
    toggleSecondSwitchesSequential(boxes);
    t.join();
}



////////////////////////////////////////////////////////////////
/// Original switchbox, for reading purpose only
////////////////////////////////////////////////////////////////

class OriginalSwitchBox
{
public:
    OriginalSwitchBox()
        : m_firstSwitchOn(true),
          m_secondSwitchOn(true)
    {
    }

    bool isFirstSwitchOn() const { return m_firstSwitchOn; }
    bool isSecondSwitchOn() const { return m_secondSwitchOn; }

    void toggleFirstSwitch() { m_firstSwitchOn = !m_firstSwitchOn; }
    void toggleSecondSwitch() { m_secondSwitchOn = !m_secondSwitchOn; }

private:
    bool m_firstSwitchOn : 1,
         m_secondSwitchOn : 1;
};



////////////////////////////////////////////////////////////////
/// Your own improved switchbox, should be modified
////////////////////////////////////////////////////////////////

class ImprovedSwitchBox
{
public:
    ImprovedSwitchBox()
        : m_firstSwitchOn(true),
          m_secondSwitchOn(true)
    {
    }

    bool isFirstSwitchOn() const { return m_firstSwitchOn; }
    bool isSecondSwitchOn() const { return m_secondSwitchOn; }

    void toggleFirstSwitch() { m_firstSwitchOn = !m_firstSwitchOn; }
    void toggleSecondSwitch() { m_secondSwitchOn = !m_secondSwitchOn; }

private:
    bool m_firstSwitchOn alignas(64);
    bool m_secondSwitchOn alignas(64);
};



int main()
{
    executeTest<OriginalSwitchBox>(toggleAllSwitchesSequential<OriginalSwitchBox>, "original switches sequential");
    executeTest<OriginalSwitchBox>(toggleAllSwitchesParallel<OriginalSwitchBox>, "original switches parallel");
    executeTest<ImprovedSwitchBox>(toggleAllSwitchesParallel<ImprovedSwitchBox>, "improved switches parallel");
    return 0;
}
